//
//  main.c
//  pgsql_test
//
//  Created by Marco Bambini on 10/02/15.
//  Copyright (c) 2015 SQLabs. All rights reserved.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef CUBESQL_WINDOWS
#define LIBRARY_PATH					"postgresql\\libpq.dll"
#endif

#ifdef CUBESQL_LINUX
#define LIBRARY_PATH					"postgresql/libpq.so"
#endif

#ifdef CUBESQL_MAC
#define LIBRARY_PATH					"postgresql/libpq.dylib"
#endif

#ifdef WIN32
#else
#include <dlfcn.h>
#endif

#include "engines.h"
#include "pg_bridge.h"

#define DEFAULTSCHEMA					"public"
#define ENGINE_NAME						"PostgreSQL"
#define ENGINE_VERSION					"1.0.1"
#define ENGINE_COPYRIGHT				"(c) 2015 SQLabs"
#define ENGINE_DESCRIPTION				"Official PostgreSQL database engine for cubeSQL"

#pragma mark -

static char	*pg_names[]					= {	"PQoidValue", "PQexec", "PQconnectdb", "PQclear", "PQstatus", "PQgetResult", "PQfinish",
											"PQerrorMessage", "PQfname", "PQntuples", "PQresultStatus", "PQgetvalue", "PQsendQuery",
											"PQcmdTuples", "PQgetlength", "PQftype", "PQnfields", "PQgetisnull", NULL};
static void *pg_func[sizeof(pg_names)/sizeof(char *)];

static int query_columntype (Oid type) {
	switch (type) {
			// MKDBTYPE_INTEGER
		case BOOLOID:
		case INT8OID:
		case INT2OID:
		case INT4OID:
		case REGPROCOID:
		case OIDOID:
		case XIDOID:
		case CIDOID:
			return COLTYPE_INTEGER;
			break;
			
			// MKDBTYPE_DATETIME
		case ABSTIMEOID:
		case RELTIMEOID:
		case TINTERVALOID:
		case DATEOID:
		case TIMEOID:
		case TIMESTAMPOID:
		case TIMESTAMPTZOID:
		case INTERVALOID:
		case TIMETZOID:
			return COLTYPE_TIMESTAMP;
			break;
			
			// MKDBTYPE_DOUBLE
		case FLOAT4OID:
		case FLOAT8OID:
		case NUMERICOID:
			return COLTYPE_FLOAT;
			break;
			
			// MKDBTYPE_BLOB
		case BYTEAOID:
			return COLTYPE_BLOB;
			break;
			
			// MKDBTYPE_TEXT
		default:
			return COLTYPE_TEXT;
			break;
	}
	
	return COLTYPE_TEXT;
}

CUBESQL_API void cubesql_engine_info (char **name, char **version, char **copyright, char **description) {
	*name = ENGINE_NAME;
	*version = ENGINE_VERSION;
	*copyright = ENGINE_COPYRIGHT;
	*description = ENGINE_DESCRIPTION;
}

CUBESQL_API void *cubesql_engine_connect (void *client, const char *dataconnection) {
	// dataconnection is for example 'host=cubesql.com port=5432 dbname=testdb user=sqlabs password=sqlabs'
	// valid fields are listed on: http://www.postgresql.org/docs/8.1/static/libpq.html
	PGconn *db = PQconnectdb(dataconnection);
	if ((!db) || (PQstatus(db) != CONNECTION_OK)) goto error_connect;
	return (void*)db;
	
error_connect:
	client_set_error(client, kExternalEngineError, PQerrorMessage(db));
	client_senderror(client);
	return NULL;
}

CUBESQL_API int cubesql_engine_close (void *client) {
	PGconn *db = (PGconn*)client_getdb(client);
	if (db == NULL) return kTRUE;
	if (db) PQfinish(db);
	return kTRUE;
}

CUBESQL_API int cubesql_engine_execute (void *client, const char *sql) {
	PGconn			*db = (PGconn*)client_getdb(client);
	PGresult		*res = NULL;
	ExecStatusType	err;
	
	// send sql command to the server
	res = PQexec(db, sql);
	if (res == NULL) err = PGRES_FATAL_ERROR;
	else err = PQresultStatus(res);
	if (err != PGRES_COMMAND_OK) goto error_exec;
	
	client_set_affectedrows(client, (int)atoi(PQcmdTuples(res)));
	client_set_lastrowid(client, (int)PQoidValue(res));
	
	if (res != NULL) PQclear(res);
	return kTRUE;
	
error_exec:
	if (res != NULL) PQclear(res);
	client_set_error(client, kExternalEngineError, PQerrorMessage(db));
	return kFALSE;
}

CUBESQL_API int cubesql_engine_select (void *client, const char *sql) {
	PGconn			*db = (PGconn*)client_getdb(client);
	PGresult		*res = NULL;
	void			*c = NULL;
	char			*r = NULL;
	int				nrows = 0;
	int				ncols = 0;
	int				i, j, len, field_size;
	
	// send asynchronous query
	if (!PQsendQuery(db, sql)) goto error_select;
	
	while ((res = PQgetResult(db))) {
		if (c == NULL) {
			nrows = PQntuples(res);
			ncols = PQnfields(res);
			if (ncols == 0) goto error_select;
			
			// initial allocation
			c = cursor_allocate (nrows, ncols);
			if (c == NULL) goto error_memory;
			
			// first row, get column names and types
			for (i = 0; i < ncols; ++i) {
				r = PQfname(res, i);
				if (r) len = (int)strlen(r) + 1; else len = 0;
				
				// copy column name
				if (cursor_copy_char(c, r, len) == kFALSE) goto error_memory;
				
				// copy column type
				if (cursor_copy_int(c, query_columntype(PQftype(res, i))) == kFALSE) goto error_memory;
			}
		}
		
		// loop to copy data
		for (i = 0; i < nrows; ++i) {
			for (j = 0; j < ncols; ++j) {
				// retrieve values
				len = 0;
				r = NULL;
				if (!PQgetisnull(res, i, j)) {r = PQgetvalue(res, i, j); len = PQgetlength(res, i, j);}
				
				field_size = len;
				if (r == NULL) field_size = NULL_VALUE;
				
				// copy column data
				if (cursor_copy_char(c, r, len) == kFALSE) goto error_memory;
				
				// copy column size
				if (cursor_copy_int(c, field_size) == kFALSE) goto error_memory;
			}
		}
		PQclear(res);
	}
	
	if (c == NULL) goto error_select;
	client_sendcursor (client, c);
	return kTRUE;
	
error_select:
	if (c) cursor_free(c);
	while (res) {PQclear(res); res = PQgetResult(db);}
	client_set_error(client, kExternalEngineError, PQerrorMessage(db));
	return kFALSE;
	
error_memory:
	if (c) cursor_free(c);
	while (res) {PQclear(res); res = PQgetResult(db);}
	client_set_error(client, kMemError, "Unable to allocate enough memory to handle query request.");
	return kFALSE;
}

CUBESQL_API int cubesql_engine_tableinfo (void *client, const char *table) {
	char sql[1024];
	
	// ColumnName TEXT, FieldType INTEGER, IsPrimary BOOLEAN, NotNULL BOOLEAN, Length INTEGER
	snprintf(sql, sizeof(sql), "SELECT c.column_name, c.data_type, c.is_nullable, c.character_maximum_length, \
			 CASE constraint_name WHEN 'PRIMARY KEY' THEN 'YES' ELSE 'NO' END AS is_primary_key \
			 FROM information_schema.columns AS c \
			 LEFT JOIN information_schema.constraint_column_usage AS ccu ON ccu.column_name = c.column_name \
			 WHERE c.table_name = '%s' AND c.table_schema = '%s'", table, DEFAULTSCHEMA);
	return cubesql_engine_select(client, sql);
}

CUBESQL_API int cubesql_engine_tables (void *client) {
	char sql[1024];
	
	snprintf(sql, sizeof(sql), "SELECT table_name FROM information_schema.tables WHERE table_schema = '%s' ORDER BY table_name;", DEFAULTSCHEMA);
	return cubesql_engine_select(client, sql);
}

CUBESQL_API int cubesql_engine_init (void) {
	void	*libhandle = NULL;
	void	*p = NULL;
	int		idx = 0;
	
	char	dllpath[PATHTYPE_LEN];
	char	enginepath[PATHTYPE_LEN];
	
	// /Library/cubesql/engines/
	cubesql_getpath(enginepath, PATHTYPE_ENGINE);
	cubesql_pathcombine(dllpath, enginepath, LIBRARY_PATH);
	
	#ifdef WIN32
	#define LOAD_LIBRARY_SEARCH_DLL_LOAD_DIR	0x00000100
	#define LOAD_LIBRARY_SEARCH_DEFAULT_DIRS	0x00001000
	libhandle = LoadLibraryExA(dllpath, NULL, LOAD_LIBRARY_SEARCH_DLL_LOAD_DIR + LOAD_LIBRARY_SEARCH_DEFAULT_DIRS);
	#else
	libhandle = dlopen(dllpath, RTLD_LAZY);
	#endif
	if (libhandle == NULL) {
		#ifdef WIN32
		DEBUG_WRITE("Error loading DLL: %d", GetLastError());
		#else
		DEBUG_WRITE("Error loading DLL: %s", dlerror());
		#endif
		return kFALSE;
	}
	
	while (pg_names[idx]) {
		char *fname = pg_names[idx];
		#ifdef WIN32
		p = GetProcAddress (libhandle, fname);
		#else
		p = dlsym (libhandle, fname);
		#endif
		if (p == NULL) {DEBUG_WRITE("Unable to find %s", fname); return kFALSE;}
		
		pg_func[idx] = p;
		++idx;
	}
	
	return kTRUE;
}
